/****************************************************************************\
*                                                                            *
*                                   WL                                       *
*                                                                            *
* Wilderland - A Hobbit Environment                                          *
*                                                                            *
* (c) 2012 by CH. Contact: wilderland@aon.at                                 *
*                                                                            *
* Simple Direct Media Layer library (SDL 1.2) from www.libsdl.org (LGPL)     *
* Z80 emulator based on Marat Fayzullin's work from 2007 (fms.komkon.org)    *
* 8x8 character set from ZX Spectrum ROM (c) by Amstrad, PD for emulators    *
*                                                                            *
* Compiler: Pelles C for Windows 6.5.80 with 'Microsoft Extensions' enabled  *
*                                                                            *
* V 1.05 - 20121026                                                          *
*                                                                            *
\****************************************************************************/



/*** INCLUDES ***************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL/SDL.h>

#include "WL.h"
#include "SDLTWE.h"
#include "SPECTRUM.h"
#include "Z80/Z80.h"
#include "MapCoordinates.h"



/*** GLOBAL VARIABLES *******************************************************/
#define EXTERN
#include "GLOBAL_VARS.h"

extern color_t ColorTable[];    //defined in Spectrum.c



/****************************************************************************\
* InitSpectrum                                                               *
*                                                                            *
\****************************************************************************/
int InitSpectrum(void)
{
    int i;
    FILE *fp;
    long int FileLength;
    size_t BytesRead;

    for (i = 0; i < 16384; i++)
        mem[i] = 0xC9;  /* = RET, just in case... */

    // we need the characte set for the lower window
    fp = fopen(SDLTWE_CHARSETFILEPATH, "rb");
    if (!fp)
    {
        fprintf(stderr, "ERROR in WL.c: can't open character set file '%s'\n", SDLTWE_CHARSETFILEPATH);
        return (1);
    }
    fseek(fp, 0, SEEK_END);
    FileLength = ftell(fp);
    fseek(fp, 0, SEEK_SET);
    if (WL_DEBUG)
    {
        printf("WL: Reading character set with %li byte from %s ... ", FileLength, SDLTWE_CHARSETFILEPATH);
    }
    BytesRead = fread(mem + SL_CHARSET - 1, 1, FileLength, fp);
    fclose(fp);
    if (WL_DEBUG)
    {
        if (FileLength != BytesRead)
            printf("ERROR!\n");
        else
            printf("read!\n");
    }

    GameWin.PosX = GAMEWINPOSX;
    GameWin.PosY = GAMEWINPOSY;

    return (0);
}


/****************************************************************************\
* InitGame                                                                   *
*                                                                            *
\****************************************************************************/
int InitGame(int hv)
{
    char GameFileName[100];
    word GameStartAddress, GameLength;
    FILE *fp;
    word FileLength, BytesRead;

    switch (hv)
    {
        case OWN:
            strcpy(GameFileName, FN_OWN);
            GameStartAddress = STARTADR_OWN;
            GameLength = LENGTH_OWN;
            DictionaryBaseAddress = DICTIONARY_BASE_OWN;
            ObjectsIndexAddress = OBJECTS_INDEX_OWN;
            ObjectsAddress = OBJECTS_OWN;
            break;
        case V10:
            strcpy(GameFileName, FN_V10);
            GameStartAddress = STARTADR_V10;
            GameLength = LENGTH_V10;
            DictionaryBaseAddress = DICTIONARY_BASE_V10;
            ObjectsIndexAddress = OBJECTS_INDEX_V10;
            ObjectsAddress = OBJECTS_V10;
            break;
        case V12:
            strcpy(GameFileName, FN_V12);
            GameStartAddress = STARTADR_V12;
            GameLength = LENGTH_V12;
            DictionaryBaseAddress = DICTIONARY_BASE_V12;
            ObjectsIndexAddress = OBJECTS_INDEX_V12;
            ObjectsAddress = OBJECTS_V12;
            break;
        default:
            fprintf(stderr, "WL: ERROR in 'InitGame'. Unknown version %d.\n", hv);
            return (-1);
    }

    fp = fopen(GameFileName, "rb");
    if (!fp)
    {
        fprintf(stderr, "WL: ERROR in 'InitGame'. Can't open game file '%s'\n", GameFileName);
        return (-1);
    }

    fseek(fp, 0, SEEK_END);
    FileLength = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    if (WL_DEBUG)
    {
        printf("WL: Reading game file '%s' with %i byte ... \n", GameFileName, FileLength);
    }

    BytesRead = fread(mem + GameStartAddress, 1, FileLength, fp);
    fclose(fp);

    if (BytesRead != GameLength)
    {
        printf("WL: ERROR in 'InitGame'. Number of bytes read doesn't match filelength.\n");
        return (-1);
    }

    return 0;
}


/****************************************************************************\
* InitGraphicsSystem                                                         *
*                                                                            *
* The global variable  SDL_Surface *screen  must be defined!                 *
\****************************************************************************/
int InitGraphicsSystem(Uint32 WinMode)
{
    SDL_Surface *temp;

    if (SDL_Init(SDL_INIT_VIDEO) < 0)
        return 1;

    if (!(screen = SDL_SetVideoMode(MAINWINWIDTH, MAINWINHEIGTH, COLORDEPTH, WinMode | SDL_HWSURFACE)))
    {
        SDL_Quit();
        return 1;
    }

    CharSet.Bitmap = malloc(SDLTWE_CHARSETLENGTH);
    if (!(CharSet.Bitmap))
    {
        fprintf(stderr, "WL: ERROR in allocationg memory for character set.\n");
        return 1;
    }
    if (!SDLTWE_ReadCharSet(SDLTWE_CHARSETFILEPATH, CharSet.Bitmap))
    {
        fprintf(stderr, "WL: ERROR while reading character set.\n");
        return 1;
    }

    // Load Game Map
    temp = SDL_LoadBMP("GameMap.bmp");
    if (temp == NULL)
    {
        printf("WL: ERROR in 'InitGraphicsSystem'. Unable to load GameMap.bmp\n");
        return (-1);
    }
    GameMap = SDL_DisplayFormat(temp);
    SDL_FreeSurface(temp);


    CharSet.Height = 8;
    CharSet.Width = 8;
    CharSet.CharMin = '\x20';
    CharSet.CharMax = '\x7F';
    CharSet.CharSubstitute = '?';

    GameWin.TW_SDL_Surface = screen;    //screen is the global SDL surface
    GameWin.PosX = GAMEWINPOSX;
    GameWin.PosY = GAMEWINPOSY;
    GameWin.Width = GAMEWINWIDTH;
    GameWin.Heigth = GAMEWINHEIGTH;
    GameWin.CurrentPrintPosX = 0;
    GameWin.CurrentPrintPosY = 0;

    LogWin.TW_SDL_Surface = screen;    //screen is the global SDL surface
    LogWin.PosX = LOGWINPOSX;
    LogWin.PosY = LOGWINPOSY;
    LogWin.Width = LOGWINWIDTH;
    LogWin.Heigth = LOGWINHEIGTH;
    LogWin.CurrentPrintPosX = 0;
    LogWin.CurrentPrintPosY = 0;

    ObjWin.TW_SDL_Surface = screen;    //screen is the global SDL surface
    ObjWin.PosX = OBJWINPOSX;
    ObjWin.PosY = OBJWINPOSY;
    ObjWin.Width = OBJWINWIDTH;
    ObjWin.Heigth = OBJWINHEIGTH;
    ObjWin.CurrentPrintPosX = 0;
    ObjWin.CurrentPrintPosY = 0;

    HelpWin.TW_SDL_Surface = screen;    //screen is the global SDL surface
    HelpWin.PosX = HELPWINPOSX;
    HelpWin.PosY = HELPWINPOSY;
    HelpWin.Width = HELPWINWIDTH;
    HelpWin.Heigth = HELPWINHEIGTH;
    HelpWin.CurrentPrintPosX = 0;
    HelpWin.CurrentPrintPosY = 0;

    return 0;
}


/****************************************************************************\
* LockScreen                                                                 *
*                                                                            *
\****************************************************************************/
void LockScreen(SDL_Surface *LSscreen)
{
    if (LockLevel)
        return;

    LockLevel++;

    if (SDL_MUSTLOCK(LSscreen))
    {
        if (SDL_LockSurface(LSscreen) < 0)
            fprintf(stderr, "WL: ERROR in LockScreen().\n");
    }
}


/****************************************************************************\
* UnLockScreen                                                               *
*                                                                            *
\****************************************************************************/
void UnLockScreen(SDL_Surface *ULSscreen)
{

    if (LockLevel-- >= 2)
    {
        SDL_Flip(screen);
        return;
    }

    if (LockLevel != 1)
    {
        SDL_Flip(screen);
        return;
    }

    if (SDL_MUSTLOCK(ULSscreen))
        SDL_UnlockSurface(ULSscreen);

    SDL_Flip(screen);
}


/****************************************************************************\
* Setpixel                                                                   *
*                                                                            *
\****************************************************************************/
void SetPixel(SDL_Surface *scr, int x, int y, color_t color)
{
    Uint32 *pixmem32;

    pixmem32 = (Uint32 *)scr->pixels + y * scr->pitch / (COLORDEPTH / 8) + x;
    *pixmem32 = color;
}


/****************************************************************************\
* PrintChar                                                                  *
*                                                                            *
\****************************************************************************/
void PrintChar(SDL_Surface *scr, struct CharSetStruct *cs, char a, int x, int y, color_t ink, color_t paper)
{
    int i, j;
    int CharIndex;
    byte mask;
    int ymul;
    Uint32 *pixm;

    if (WL_DEBUG)
    {
        if ((cs->Width != 8) || (cs->Height != 8))
        {
            fprintf(stderr, "WL: ERROR in PrintChar. Only 8x8 pixel charsets are supported.\n");
            return;
        }
    }


    ymul = scr->pitch / (COLORDEPTH / 8);
    for (i = 0; i <= 7; i++)
    {
        mask = 0x80;
        CharIndex = (a - cs->CharMin) * cs->Height;
        for (j = 0; j <= 7; j++, mask >>= 1)
        {
            pixm = (Uint32 *) (scr->pixels) + (y + i) * ymul + x + j;
            if (cs->Bitmap[CharIndex + i] & mask)
                *pixm = ink;
            else
                *pixm = paper;
        }
    }
}

/****************************************************************************\
* PrintString                                                                *
*                                                                            *
\****************************************************************************/
void PrintString(SDL_Surface *scr, struct CharSetStruct *cs, char *ps, int x, int y, color_t ink, color_t paper)
{
    while (*ps)
    {
        PrintChar(scr, cs, *ps++, x, y, ink, paper);
        x += cs->Width;
    }
}


/****************************************************************************\
* DrawLineRelative                                                           *
*                                                                            *
\****************************************************************************/
void DrawLineRelative(SDL_Surface *scr, int x, int y, int dx, int dy, int rx, int ry, color_t color)
{
    int x1, y1, sx, sy;

    if (rx == 0 || ry == 0)
        return;

    sx = (dx > 0) ? 1 : -1;
    dx *= sx;

    sy = (dy > 0) ? 1 : -1;
    dy *= sy;

    while (rx > 1 || ry > 1)
    {
        if (rx > 1)
            rx--;
        if (ry > 1)
            ry--;
        if (dx > dy)
            for (x1 = 0; x1 < dx; x1++)
                SetPixel(scr, x + sx * (x1 + (rx - 1)), sy * (y + ((int)(dy * x1 / dx)) + (ry - 1)), color);
        else
            for (y1 = 0; y1 < dy; y1++)
                SetPixel(scr, x + sx * (((int)(dx * y1 / dy)) + (rx - 1)), y + sy * (y1 + (ry - 1)), color);
    }
}


/****************************************************************************\
* Info                                                                       *
*                                                                            *
\****************************************************************************/
void Info(struct TextWindowStruct *TW, struct CharSetStruct *CS)
{
    LockScreen(TW->TW_SDL_Surface);

    SDLTWE_PrintString(TW, "\nWILDERLAND - A Hobbit Environment (c) 2012 by CH\n\n", CS, SC_BRWHITE, SC_BRBLACK);
    SDLTWE_PrintString(TW, "\"The Hobbit\" (c) Melbourne House, 1982. Written by Philip\n", CS, SC_BRMAGENTA, SC_BRBLACK);
    SDLTWE_PrintString(TW, "Mitchell and Veronika Megler.\n\n", CS, SC_BRMAGENTA, SC_BRBLACK);
    SDLTWE_PrintString(TW, "Simple Direct Media Layer library (SDL 1.2) from www.libsdl.org\n", CS, SC_WHITE, SC_BLACK);
    SDLTWE_PrintString(TW, "Z80 emulator based on Marat Fayzullin's work (fms.komkon.org)\n", CS, SC_WHITE, SC_BLACK);
    SDLTWE_PrintString(TW, "8x8 charset from ZX Spectrum ROM (c) by Amstrad,PD for emulators\n\n", CS, SC_WHITE, SC_BLACK);
    SDLTWE_PrintString(TW, "You are not allowed to distribute 'WILDERLAND' with Hobbit\nbinaries included!\n\n", CS, SC_BRWHITE, SC_BRBLACK);
    SDLTWE_PrintString(TW, "Contact: wilderland@aon.at\n", CS, SC_WHITE, SC_BLACK);

    UnLockScreen(TW->TW_SDL_Surface);
}

/****************************************************************************\
* HelpInfo                                                                   *
*                                                                            *
\****************************************************************************/
void HelpInfo(struct TextWindowStruct *TW, struct CharSetStruct *CS)
{
    LockScreen(TW->TW_SDL_Surface);

    SDLTWE_PrintString(TW, "\n                         WILDERLAND V 1.5\n\n", CS, SC_BRWHITE, SC_BRBLACK);
    SDLTWE_PrintString(TW, "                       A Hobbit Environment\n\n", CS, SC_BRWHITE, SC_BRBLACK);
    SDLTWE_PrintString(TW, "                          (c) 2012 by CH\n\n\n", CS, SC_WHITE, SC_BLACK);
    SDLTWE_PrintString(TW, " a-z.,@0[BS] -> play         ATTRIBUTES\n", CS, SC_BRMAGENTA, SC_BRBLACK);
    SDLTWE_PrintString(TW, " S-ave   H-elp   G-o         v_isible  A_nimal  o_pen   *_light\n", CS, SC_BRCYAN, SC_BRBLACK);
    SDLTWE_PrintString(TW, " L-oad   I-nfo               b_roken   f_ull    F_luid  l_ocked\n", CS, SC_BRCYAN, SC_BRBLACK);
    SDLTWE_PrintString(TW, " Q-uit\n\n", CS, SC_BRWHITE, SC_BRBLACK);
    SDLTWE_PrintString(TW, " Press 'n' as first key to play without graphics (not game 1.0).\n", CS, SC_RED, SC_BLACK);

    UnLockScreen(TW->TW_SDL_Surface);
}


/****************************************************************************\
* DisplayGameMap                                                             *
*                                                                            *
\****************************************************************************/
void DisplayGameMap(void)
{
    SDL_Rect src, dest;

    src.x = 0;
    src.y = 0;
    src.w = GAMEMAPWIDTH;
    src.h = GAMEMAPHEIGTH;

    dest.x = 0;
    dest.y = (MAINWINHEIGTH - GAMEMAPHEIGTH - 1);
    dest.w = GAMEMAPWIDTH;
    dest.h = GAMEMAPHEIGTH;

    SDL_BlitSurface(GameMap, &src, screen, &dest);
}


/****************************************************************************\
* GetHexByte                                                                 *
*                                                                            *
\****************************************************************************/
void GetHexByte(byte *b, struct TextWindowStruct *TW, struct CharSetStruct *CS, color_t ink, color_t paper)
{
    Uint16 a;
    SDL_Event event;
    int nibble_count = 0;

    *b = 0;

    while (1)
    {
        if (SDL_PollEvent(&event))
        {
            switch (event.type)
            {
                case SDL_KEYDOWN:
                    a = event.key.keysym.unicode;
                    if (a >= '0' && a <= '9')
                    {
                        if (nibble_count >= 2) break;
                        *b <<= 4;
                        *b += (a - '0');
                        nibble_count++;
                        LockScreen(screen);
                        SDLTWE_PrintCharTextWindow(TW, a, CS, ink, paper);
                        UnLockScreen(screen);
                        break;
                    }
                    if ((a >= 'a' && a <= 'f') || (a >= 'A' && a <= 'F'))
                    {
                        if (nibble_count >= 2) break;
                        a |= 0x20;
                        *b <<= 4;
                        *b += (a - 'a' + 0x0a);
                        nibble_count++;
                        LockScreen(screen);
                        SDLTWE_PrintCharTextWindow(TW, a, CS, ink, paper);
                        UnLockScreen(screen);
                        break;
                    }
                    if (a == SDLK_BACKSPACE)
                        break;
                    if (a == SDLK_RETURN)
                    {
                        return;
                    }
                case SDL_KEYUP:
                    a = 0;
                    break;
            }
        }
    }
}


/****************************************************************************\
* GetFileName                                                                *
*                                                                            *
\****************************************************************************/
void GetFileName(char *fnstr, struct TextWindowStruct *TW, struct CharSetStruct *CS, color_t ink, color_t paper)
{
    Uint16 a;
    SDL_Event event;

    LockScreen(screen);
    SDLTWE_PrintString(TW, "\n\nFilename: ", CS, ink, paper);
    UnLockScreen(screen);

    while (1)
    {
        if (SDL_PollEvent(&event))
        {
            switch (event.type)
            {
                case SDL_KEYDOWN:
                    a = event.key.keysym.unicode;
                    if ((a >= 'a' && a <= 'z') || (a >= '0' && a <= '9') || a == '.' || a == '_')
                    {
                        *fnstr = a;
                        fnstr++;
                        LockScreen(screen);
                        SDLTWE_PrintCharTextWindow(TW, a, CS, ink, paper);
                        UnLockScreen(screen);
                        break;
                    }
                    if (a == SDLK_BACKSPACE)
                        break;
                    if (a == SDLK_RETURN)
                    {
                        *fnstr = 0;
                        return;
                    }
                case SDL_KEYUP:
                    a = 0;
                    break;
            }
        }
    }
}


/****************************************************************************\
* SaveGame                                                                   *
*                                                                            *
\****************************************************************************/
void SaveGame(struct TextWindowStruct *TW, struct CharSetStruct *CS, int hv, color_t ink, color_t paper)
{
    char fn[100];
    FILE *f;
    int i;

    GetFileName(fn, TW, CS, ink, paper);

    f = fopen(fn, "wb");
    if (!f)
    {
        fprintf(stderr, "WL: ERROR - Can't open file '%s' for saving game.\n", fn);
        return;
    }

    fprintf(f, "GAMEVERSION=%i\r\n", hv);
    fprintf(f, "FILEVERSION=1\r\n");
    fprintf(f, "SIZE(Z80)=%02zX\r\n", sizeof(z80));
    fprintf(f, "PC=%04X\r\n", z80.PC.W);

    fprintf(f, "***Z80***");
    for (i = 0; i < sizeof(z80); i++)
        fprintf(f, "%c", *((byte *) & z80 + i));

    fprintf(f, "***MEMORY***");
    for (i = 0; i <= 0xFFFF; i++)
        fprintf(f, "%c", mem[i]);

    fclose(f);

    LockScreen(screen);
    SDLTWE_PrintString(TW, " saved.\n\n", CS, ink, paper);
    UnLockScreen(screen);
}


/****************************************************************************\
* LoadGame                                                                   *
*                                                                            *
\****************************************************************************/
void LoadGame(struct TextWindowStruct *TW, struct CharSetStruct *CS, int hv, color_t ink, color_t paper)
{
    char fn[100];
    FILE *f;
    int i, gameversion, fileversion, SizeOfZ80;

    GetFileName(fn, TW, CS, ink, paper);

    f = fopen(fn, "rb");
    if (!f)
    {
        fprintf(stderr, "WL: ERROR - Can't open file '%s' for loading game.\n", fn);
        LockScreen(screen);
        SDLTWE_PrintString(TW, " not found!\n\n", CS, ink, paper);
        UnLockScreen(screen);
        return;
    }

    if (fscanf(f, "GAMEVERSION=%i\r\n", &gameversion) != 1)
    {
        LockScreen(screen);
        SDLTWE_PrintString(TW, "\nError. Data does not match format.\n\n", CS, ink, paper);
        UnLockScreen(screen);
        return;
    }

    if (gameversion != hv)
    {
        LockScreen(screen);
        SDLTWE_PrintString(TW, "\nWrong game version.\n\n", CS, ink, paper);
        UnLockScreen(screen);
        return;
    }

    if (fscanf(f, "FILEVERSION=%i\r\n", &fileversion) != 1)
    {
       LockScreen(screen);
       SDLTWE_PrintString(TW, "\nError. File version does not match format.\n\n", CS, ink, paper);
       UnLockScreen(screen);
       return;
    }

    if (fscanf(f, "SIZE(Z80)=%02X\r\n", &SizeOfZ80) != 1)
    {
       LockScreen(screen);
       SDLTWE_PrintString(TW, "\nError. SIZE(Z80) does not match format.\n\n", CS, ink, paper);
       UnLockScreen(screen);
       return;
    }

    if (SizeOfZ80 != sizeof(z80))
    {
        LockScreen(screen);
        SDLTWE_PrintString(TW, "\nZ80-structure sizes do not match.\n\n", CS, ink, paper);
        UnLockScreen(screen);
        return;
    }

    if (fileversion == 1)
    {
        if (fscanf(f, "PC=%04X\r\n", &fileversion) != 1)
        {
           LockScreen(screen);
           SDLTWE_PrintString(TW, "\nError. Fileversion does not match format.\n\n", CS, ink, paper);
           UnLockScreen(screen);
           return;
        }
        if (fscanf(f, "***Z80***") != 1)
        {
           LockScreen(screen);
           SDLTWE_PrintString(TW, "\nError. ***Z80*** does not match format.\n\n", CS, ink, paper);
           UnLockScreen(screen);
           return;
        }
        for (i = 0; i < sizeof(z80); i++)
            if (fscanf(f, "%c", ((byte *) & z80 + i)) != 1)
            {
               LockScreen(screen);
               SDLTWE_PrintString(TW, "\nError. %%c does not match format.\n\n", CS, ink, paper);
               UnLockScreen(screen);
               return;
            }
        if (fscanf(f, "***MEMORY***") != 1)
        {
           LockScreen(screen);
           SDLTWE_PrintString(TW, "\nError. ***MEMORY*** does not match format.\n\n", CS, ink, paper);
           UnLockScreen(screen);
           return;
        }
        for (i = 0; i <= 0xFFFF; i++)
            if (fscanf(f, "%c", mem + i) != 1)
            {
               LockScreen(screen);
               SDLTWE_PrintString(TW, "\nError. %%c does not match format.\n\n", CS, ink, paper);
               UnLockScreen(screen);
               return;
            }
        fclose(f);
    }
    else
    {
        LockScreen(screen);
        SDLTWE_PrintString(TW, "\nThis game file version is not supported!\n\n", CS, ink, paper);
        UnLockScreen(screen);
        return;
    }

    LockScreen(screen);
    SDLTWE_PrintString(TW, " loaded.\n\n", CS, ink, paper);
    UnLockScreen(screen);

    for (i = SL_SCREENSTART; i <= SL_ATTRIBUTEEND; i++)
        WrZ80(i, mem[i]);
}


/****************************************************************************\
* Go                                                                         *
*                                                                            *
\****************************************************************************/
void Go(struct TextWindowStruct *TW, struct CharSetStruct *CS, int hv, color_t ink, color_t paper)
{
    byte rn;

    LockScreen(screen);
    SDLTWE_PrintString(TW, "\n\nEnter room number 0x", CS, ink, paper);
    UnLockScreen(screen);

    GetHexByte(&rn, TW, CS, ink, paper);

    if (rn < 1 || rn > ROOMS_NR_MAX)
    {
        LockScreen(screen);
        SDLTWE_PrintString(TW, ". ERROR, invalid room number.\n", CS, ink, paper);
        UnLockScreen(screen);
        return;
    }

    switch (hv)
    {
        case OWN:
            mem[L_YOU_OBJ_POSITION_OWN] = rn;
            break;
        case V10:
            mem[L_YOU_OBJ_POSITION_V10] = rn;
            break;
        case V12:
            mem[L_YOU_OBJ_POSITION_V12] = rn;
            break;
        default:
            fprintf(stderr, "ERROR in WL/Go: unknown game version.\n");
            return;
    }

    LockScreen(screen);
    SDLTWE_PrintString(TW, ". OK, you are now in this room.\n", CS, ink, paper);
    UnLockScreen(screen);
}


/****************************************************************************\
* sbinprintf                                                                 *
*                                                                            *
\****************************************************************************/
void sbinprintf(char *buf, byte b)
{
    int i;
    byte mask = 0x80;

    for (i = 0; i < 8; i++, mask >>= 1)
    {
        *buf++ = '0' + ((b & mask) != 0);
        if (i == 3)
            *buf++ = '.';
    }
    *buf = 0;
}


/****************************************************************************\
* SearchByteWordTable                                                        *
*                                                                            *
* Returns pointer value or 0 if not found.                                   *
\****************************************************************************/
word SearchByteWordTable(byte a, word address)
{
    while (mem[address] != 0xFF)
    {
        if (mem[address] == a)
            return (mem[address + 1] + 0x100 * mem[address + 2]);
        else
            address += 3;
    }
    return (0);
}


/****************************************************************************\
* GetDictWord                                                                *
*                                                                            *
\****************************************************************************/
void GetDictWord(word a, char *buf)
{
    word i = 1; // char count

    a += DictionaryBaseAddress;

    // the following line relies on evaluation from left to right
    while (!(mem[a] & 0x80) || (i == 2) || ((i == 3) && (mem[a - 1] & 0x80)))
    {
        if (mem[a] & 0x1F)
        {
            *buf++ = (mem[a] & 0x1F) + 'A' - 1;
        }
        a++;
        i++;
    }
    if (mem[a] & 0x1F)
        *buf++ = (mem[a] & 0x1F) + 'A' - 1;
    *buf = 0;
}


/****************************************************************************\
* GetObjectAttributePointer                                                  *
*                                                                            *
* Returns pointer value or 0 if not found.                                   *
\****************************************************************************/
word GetObjectAttributePointer(byte a, byte attributeoffset)
{
    word OAP;

    OAP = SearchByteWordTable(a, ObjectsIndexAddress);
    if (OAP)
        return (OAP + attributeoffset);
    else
        return (0);
}


/****************************************************************************\
* GetObjectFullName                                                          *
*                                                                            *
* +0A/0B: ADJECTIVE1, +0C/0D: ADJECTIVE2, +08/09: NOUN                       *
\****************************************************************************/
void GetObjectFullName(word oa, char *OFN)
{
    word wa;
    char StringBuffer[25];

    wa = mem[oa + 0x0A] + 0x100 * mem[oa + 0x0B];   // adjective 1
    if (wa)
    {
        GetDictWord(wa, StringBuffer);
        strcat(OFN, StringBuffer);
        strcat(OFN, " ");
    }
    wa = mem[oa + 0x0C] + 0x100 * mem[oa + 0x0D];   // adjective 2
    if (wa)
    {
        GetDictWord(wa, StringBuffer);
        if (!strcmp(StringBuffer, "INSIGNIFICANT"))
            strcpy(StringBuffer, "INSIGN.");
        strcat(OFN, StringBuffer);
        strcat(OFN, " ");
    }
    wa = mem[oa + 0x08] + 0x100 * mem[oa + 0x09];   // noun
    if (wa & 0x0FFF)
    {
        GetDictWord(wa & 0x0FFF, StringBuffer);
        strcat(OFN, StringBuffer);
    }

}


/****************************************************************************\
* DrawAnimalPositon                                                          *
*                                                                            *
\****************************************************************************/
void DrawAnimalPosition(SDL_Surface *scr, struct CharSetStruct *cs, byte animalnr, word x, word y, byte objectoffset)
{
    char initials[10];

    DrawLineRelative(scr, x, y + objectoffset*15, 10, -30, 4, 1, SC_BRRED);

    strcpy(initials, AnimalInitials[animalnr]);
    if (mem[GetObjectAttributePointer(animalnr, ATTRIBUTE_OFF)] & ATTR_BROKEN)
        strcat(initials, "+");
    if (animalnr == OBJNR_YOU || animalnr == OBJNR_GANDALF || animalnr == OBJNR_THORIN)
        PrintString(scr, cs, initials, x + 13, y + objectoffset*15 - 33, SC_BRRED, 0x00000000ul);
    else
        PrintString(scr, cs, initials, x + 13, y + objectoffset*15 - 33, SC_BRGREEN, 0x00000000ul);
}


/****************************************************************************\
* DisplayOneAnimalPositon                                                    *
*                                                                            *
\****************************************************************************/
void DisplayOneAnimalPosition(SDL_Surface *scr, struct CharSetStruct *cs, byte AnimalNr, byte *objectsperroom)
{
    byte CurrentAnimalRoom;
    int i;

    CurrentAnimalRoom = mem[GetObjectAttributePointer(AnimalNr, FIRST_OCCURRENCE)];

    for(i = 0; i<ROOMS_NROF_MAX; i++)
    {
        if (MapCoordinates[i].RoomNumber == CurrentAnimalRoom)
        {
            DrawAnimalPosition(scr, cs, AnimalNr, MapCoordinates[i].X - INDICATOROFFSET,
                MapCoordinates[i].Y + (MAINWINHEIGTH - GAMEMAPHEIGTH) + INDICATOROFFSET,
                objectsperroom[CurrentAnimalRoom]);
            objectsperroom[CurrentAnimalRoom]++;

        }
    }
}


/****************************************************************************\
* DisplayAnimalPositions                                                     *
*                                                                            *
\****************************************************************************/
void DisplayAnimalPositions(SDL_Surface *scr, struct CharSetStruct *cs, byte *objectsperroom)
{
    byte AnimalNr;

    DisplayOneAnimalPosition(scr, cs, OBJNR_YOU, objectsperroom);

    for(AnimalNr = OBJNR_DRAGON; AnimalNr <= OBJNR_DISGUSTINGGOBLIN; AnimalNr++)
            DisplayOneAnimalPosition(scr, cs, AnimalNr, objectsperroom);
}


/****************************************************************************\
* PrintObject                                                                *
*                                                                            *
\****************************************************************************/
void PrintObject(word ai, struct TextWindowStruct *OW, struct CharSetStruct *CS, color_t ink, color_t paper)
{
    word oa;
    byte ObjectNumber;
    char ObjectPrintLine[100];
    char ObjectStringBuffer[100];

    ObjectNumber = mem[ai++];
    sprintf(ObjectPrintLine, "%02X ", ObjectNumber);

    oa = mem[ai] + 0x100 * mem[ai + 1];
    // object name
    *ObjectStringBuffer = 0;
    GetObjectFullName(oa, ObjectStringBuffer);
    strcat(ObjectStringBuffer, "                              ");
    ObjectStringBuffer[21] = 0;
    strcat(ObjectPrintLine, ObjectStringBuffer);
    // 1st occurence
    sprintf(ObjectStringBuffer, "%02X ", mem[oa + FIRST_OCCURRENCE]);
    strcat(ObjectPrintLine, ObjectStringBuffer);
    // occurences
    sprintf(ObjectStringBuffer, "%1X ", mem[oa + OCC_OFF]);
    strcat(ObjectPrintLine, ObjectStringBuffer);
    // mother object
    sprintf(ObjectStringBuffer, "%02X ", mem[oa + MO_OFF]);
    strcat(ObjectPrintLine, ObjectStringBuffer);
    // volume
    sprintf(ObjectStringBuffer, "%02X ", mem[oa + VOLUME_OFF]);
    strcat(ObjectPrintLine, ObjectStringBuffer);
    // mass
    sprintf(ObjectStringBuffer, "%02X ", mem[oa + MASS_OFF]);
    strcat(ObjectPrintLine, ObjectStringBuffer);
    // +04
    sprintf(ObjectStringBuffer, "%02X ", mem[oa + A4_OFF]);
    strcat(ObjectPrintLine, ObjectStringBuffer);
    // +05
    sprintf(ObjectStringBuffer, "%02X ", mem[oa + A5_OFF]);
    strcat(ObjectPrintLine, ObjectStringBuffer);
    // +06
    sprintf(ObjectStringBuffer, "%02X ", mem[oa + A6_OFF]);
    strcat(ObjectPrintLine, ObjectStringBuffer);
    // ATTRIBUTE
    sbinprintf(ObjectStringBuffer, mem[oa + ATTRIBUTE_OFF]);
    strcat(ObjectPrintLine, ObjectStringBuffer);

    if (ObjectNumber == 0x00 || ObjectNumber == 0x3E || ObjectNumber == 0x3F)
        SDLTWE_PrintString(OW, ObjectPrintLine, CS, SC_BRRED, paper);
    else
    {
        if (mem[oa + FIRST_OCCURRENCE] == mem[GetObjectAttributePointer(OBJNR_YOU, FIRST_OCCURRENCE)])
            SDLTWE_PrintString(OW, ObjectPrintLine, CS, paper, ink);
        else
            SDLTWE_PrintString(OW, ObjectPrintLine, CS, ink, paper);
    }
}


/****************************************************************************\
* PrintObjectsList                                                           *
*                                                                            *
\****************************************************************************/
void PrintObjectsList(word OIA, word OA, struct TextWindowStruct *OW, struct CharSetStruct *CS, color_t ink, color_t paper)
{
    word ai;
    OW->CurrentPrintPosX = 0;
    OW->CurrentPrintPosY = 0;
    SDLTWE_PrintString(OW, "Nr Name                 1L # MO Vo Ma 04 St 06 vAo*.bfFl", CS, ink, paper);
    SDLTWE_PrintString(OW, "--------------------------------------------------------", CS, ink, paper);
    ai = OIA;
    while (mem[ai] != 0xFF)
    {
        if (mem[ai] == 60)  // animals have object numbers >= 60
            SDLTWE_PrintString(OW, "--------------------------------------------------------", CS, ink, paper);
        PrintObject(ai, OW, CS, ink, paper - (mem[ai] % 2 ? 0x00000040ul : 0));
        ai += 3;
    }
}


/****************************************************************************\
*                                   main                                     *
*                                                                            *
\****************************************************************************/
int main(int argc, char *argv[])
{
    SDL_Event event;
    int i;
    int RunMainLoop;

    byte b;
    FILE *f;
    Uint32 WinMode = 0; // window mode (vs. full screen);
    int SeedRND    = 0;
    int MaxSpeed   = 0;
    Uint32 msTimer = 0;
    Sint32 DeltaT  = 0;
    int FrameCount = 0;

    static byte ObjectsPerRoom[ROOMS_NROF_MAX];

    HV = -1;  // Hobbit version (global variable)

    // Help line
    if (argc == 1)
    {
        printf("Use: WL -V10|-OWN|-V12 [-FULLSCREEN] [-MAXSPEED] [-NOSCANLINES] [-SEEDRND]\n");
        exit(-1);
    }

    // process command line arguments
    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-own") || !strcmp(argv[i], "-OWN"))
            HV = OWN;
        if (!strcmp(argv[i], "-v10") || !strcmp(argv[i], "-V10"))
            HV = V10;
        if (!strcmp(argv[i], "-v12") || !strcmp(argv[i], "-V12"))
            HV = V12;
        if (!strcmp(argv[i], "-fullscreen") || !strcmp(argv[i], "-FULLSCREEN"))
            WinMode = SDL_FULLSCREEN;
        if (!strcmp(argv[i], "-maxspeed") || !strcmp(argv[i], "-MAXSPEED"))
            MaxSpeed = 1;
        if (!strcmp(argv[i], "-noscanlines") || !strcmp(argv[i], "-NOSCANLINES"))
            NoScanLines = 1;
        if (!strcmp(argv[i], "-seedrnd") || !strcmp(argv[i], "-SEEDRND"))
            SeedRND = 1;
    }


    if (HV == -1)
    {
        fprintf(stderr, "WL: ERROR Unknown game version. Program aborted.\n");
        exit(-1);
    }

    if (InitSpectrum())
    {
        fprintf(stderr, "WL: ERROR initializing spectrum. Program aborted.\n");
        exit(-1);
    }

    if (InitGame(HV))
    {
        fprintf(stderr, "WL: ERROR initializing game. Program aborted.\n");
        exit(-1);
    }

    if (InitGraphicsSystem(WinMode))
    {
        fprintf(stderr, "WL: ERROR initializing graphic system. Program aborted.\n");
        exit(-1);
    }


    LockScreen(screen);

    SDLTWE_DrawTextWindowFrame(&GameWin, 4, 0x00404040ul, 4, SC_CYAN);
    SDLTWE_DrawTextWindowFrame(&ObjWin,  4, 0x00404040ul, 4, SC_CYAN);
    SDLTWE_DrawTextWindowFrame(&HelpWin, 4, 0x00404040ul, 4, SC_CYAN);
    SDLTWE_DrawTextWindowFrame(&LogWin,  4, 0x00404040ul, 4, SC_CYAN);  //must be last to overdraw other frames
    SDLTWE_PrintCharTextWindow(&LogWin, '\f', &CharSet, SC_BLACK, SC_WHITE);


    // title screen
    f = fopen("HOBBIT.SCR", "rb");
    if (!f)
    {
        fprintf(stderr, "WL: ERROR - Can't open title screen file 'HOBBIT.SCR'\n");
        exit(-1);
    }
    else
    {
        for (i = 16384; i < ((16384 + 256 / 8 * 192) + (32 * 24)); i++)
        {
            if (fscanf(f, "%c", &b) != 1)
            {
               fprintf(stderr, "WL: Error - %%c does not match format\n");
               fclose(f);
               exit(-1);
            }
            WrZ80(i, b);
        }
        fclose(f);
    }

    HelpInfo(&HelpWin, &CharSet);

    UnLockScreen(screen);


    ResetZ80(&z80);
    switch (HV)
    {
        case OWN:
            z80.PC.W = L_START_OWN;
            break;
        case V10:
            z80.PC.W = L_START_V10;
            break;
        case V12:
            z80.PC.W = L_START_V12;
            break;
        default:
            exit(-1);
    }
    if (SeedRND)
        z80.R = (byte) SDL_GetTicks();

    SDL_EnableUNICODE(1);
    RunMainLoop = 1;



    /****************************** MAIN LOOP ******************************/
    while (RunMainLoop)
    {

        // we run the main loop with FPS Hz
        DeltaT = SDL_GetTicks() - msTimer;  // on my Intel Core2 Duo E6850@3GHz ca. 5 ms
        if (DeltaT < DELAY_MS)
            if (!MaxSpeed)
                SDL_Delay(DELAY_MS - DeltaT);
        msTimer = SDL_GetTicks();
        FrameCount++;



        LockScreen(screen);

        ExecZ80(&z80, TSTATES_PER_LOOP);

        PrintObjectsList(ObjectsIndexAddress, ObjectsAddress, &ObjWin, &CharSet, SC_BRYELLOW, SC_BRBLUE);

        if (FrameCount > FPS)
        {
            FrameCount = 0;

            for (i = 0; i < ROOMS_NROF_MAX; i++)
                ObjectsPerRoom[i] = 0;
            DisplayGameMap();
            DisplayAnimalPositions(screen, &CharSet, ObjectsPerRoom);
        }

        UnLockScreen(screen);



        if (SDL_PollEvent(&event))
        {
            switch (event.type)
            {
                case SDL_QUIT:
                    RunMainLoop = 0;
                    break;
                case SDL_KEYDOWN:
                    CurrentPressedKey = event.key.keysym.unicode;
                    if (CurrentPressedKey == SDLK_BACKSPACE)
                        CurrentPressedKey = '0';
                    if (CurrentPressedKey == 'Q')
                        RunMainLoop = 0;
                    if (CurrentPressedKey == 'S')
                        SaveGame(&HelpWin, &CharSet, HV, SC_BRGREEN, SC_BRBLACK);
                    if (CurrentPressedKey == 'L')
                        LoadGame(&HelpWin, &CharSet, HV, SC_BRGREEN, SC_BRBLACK);
                    if (CurrentPressedKey == 'H')
                        HelpInfo(&HelpWin, &CharSet);
                    if (CurrentPressedKey == 'I')
                        Info(&HelpWin, &CharSet);
                    if (CurrentPressedKey == 'G')
                        Go(&HelpWin, &CharSet, HV, SC_BRGREEN, SC_BRBLACK);
                    break;
                case SDL_KEYUP:
                    CurrentPressedKey = 0;
                    break;
            }
        }
    }

    SDL_Quit();
    exit(0);
}
